import os
from pathlib import Path

from pydantic import BaseModel

from parser.parser import parse_code


def is_python_package(path: Path):
    return (path / "__init__.py").exists()


class ModuleFile(BaseModel):
    name: str
    path: str


class DirPackageScanner:
    def __init__(self, work_path: Path):
        self.work_path = work_path
        self.packages = []
        self.files: list[ModuleFile] = []

    def add_file(self, namespace, path: Path):
        relative_path = path.relative_to(self.work_path)
        self.files.append(ModuleFile(name=namespace, path=str(relative_path)))

    def scan_python_files(self, path, namespace):
        for p in path.iterdir():
            if p.is_file() and p.suffix == ".py":
                self.add_file(namespace, p)
            elif p.is_dir():
                self.scan_python_files(p, f"{namespace}.{p.name}")

    def scan_python_package(self, path):
        namespace = path.name
        for p in path.iterdir():
            if p.is_file() and p.suffix == ".py":
                self.add_file(namespace, p)
            elif p.is_dir():
                self.scan_python_files(p, f"{namespace}.{p.name}")

    def scan_dir(self, dir_path: Path):
        for p in dir_path.iterdir():
            # if start doc or test, skip
            if p.name.startswith("doc") or p.name.startswith("test"):
                continue

            if p.is_dir():
                if is_python_package(p):
                    self.scan_python_package(p)
                else:
                    self.scan_dir(p)

    def scan(self):
        self.scan_dir(self.work_path)



class ImportsMap:
    def __init__(self):
        self.map: dict[str, list[str]] = {}

    def add_import(self, module, names: list[str]):
        if module not in self.map:
            self.map[module] = []
        self.map[module].extend(names)

    def get_libs(self):
        return self.map.keys()


class Context:
    def __init__(self, node, namespace, imports: ImportsMap):
        self.node = node
        self.namespace = namespace
        self.imports = imports


class DecoratorContext:
    def __init__(self, node, namespace):
        self.node = node
        self.namespace = namespace


def check_node(node: Node, type: str):
    if node.type != type:
        raise AssertionError(f"Expect {type}, but got {node.type}")


def to_str(node: Node):
    return node.text.decode()


def nodes_filter(nodes: List[Node], type: str):
    return [i for i in nodes if i.type == type]




from typing import List

from tree_sitter import Node

from parser.base import Symbol


class ImportsMap:
    def __init__(self):
        self.map: dict[str, list[str]] = {}

    def add_import(self, module, names: list[str]):
        if module not in self.map:
            self.map[module] = []
        self.map[module].extend(names)

    def get_libs(self):
        return self.map.keys()


class Context:
    def __init__(self, node, namespace, imports: ImportsMap):
        self.node = node
        self.namespace = namespace
        self.imports = imports


class DecoratorContext:
    def __init__(self, node, namespace):
        self.node = node
        self.namespace = namespace


def check_node(node: Node, type: str):
    if node.type != type:
        raise AssertionError(f"Expect {type}, but got {node.type}. {node.start_point}-{node.end_point}")


def to_str(node: Node):
    return node.text.decode()


def nodes_filter(nodes: List[Node], type: str):
    return [i for i in nodes if i.type == type]


def get_imports(node: Node):
    imports_map = ImportsMap()
    for child in node.children:
        match child.type:
            case 'future_import_statement':
                names = [i.text.decode() for i in child.named_children]
                imports_map.add_import("__future__", names)

            case 'import_statement':
                names = [i.text.decode() for i in child.named_children]
                for i in names:
                    imports_map.add_import(i, ["*"])

            case 'import_from_statement':
                names = [i.text.decode() for i in child.named_children]
                check_node(child.children[0], "from")
                check_node(child.children[2], "import")
                imports_map.add_import(names[0], names[1:])

    return imports_map


class PythonTreeWalker:
    def __init__(self,
                 root_node: Node,
                 content_lines: List[str],
                 module_path: str,
                 code_path: str
                 ):
        self.root_node = root_node
        self.content_lines = content_lines
        self.module_path = module_path
        self.symbols = []
        self.module_imports = []
        self.code_path = code_path

    def visit(self):
        assert self.root_node.type == 'module'
        imports = get_imports(self.root_node)
        parent = Context(self.root_node, self.module_path, imports)
        parent.namespace = self.module_path

        self.module_imports = parent.imports.get_libs()
        self.visit_block(self.root_node, parent)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L531

    def visit_decorated_definition(self, node: Node, parent: Context):
        check_node(node, "decorated_definition")

        last_child = node.children[-1]
        match last_child.type:
            case 'function_definition':
                self.visit_function_definition(last_child, parent)
            case 'class_definition':
                self.visit_class_definition(last_child, parent)
            case _:
                raise AssertionError(f"Unsupported decorated definition: {last_child.type}")

    def visit_block(self, node: Node, parent: Context):
        for child in node.children:
            match child.type:
                case 'function_definition':
                    self.visit_function_definition(child, parent)
                case 'class_definition':
                    self.visit_class_definition(child, parent)
                case 'expression_statement':
                    self.visit_expression_statement(child, parent)
                case 'decorated_definition':
                    self.visit_decorated_definition(child, parent)

    # if have decorator, will set start_point
    def visit_function_definition(self, node: Node, parent: Context, start_point=None):

        check_node(node, "function_definition")
        check_node(node.named_child(0), "identifier")

        func_name = node.children[1].text.decode()
        ctx = Context(node, f"{parent.namespace}.{func_name}", parent.imports)
        self.add_symbol("function",
                        ctx.namespace,
                        node.start_point if start_point is None else start_point,
                        node.end_point,
                        )
        self.visit_block(node, ctx)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L472
    # if have decorator, will set start_point
    def visit_class_definition(self, node: Node, parent: Context, start_point=None):
        check_node(node, "class_definition")
        check_node(node.children[0], "class")
        check_node(node.children[1], "identifier")
        check_node(node.children[-1], "block")

        class_name = node.children[1].text.decode()
        ctx = Context(node, f"{parent.namespace}.{class_name}", parent.imports)
        self.add_symbol("class",
                        ctx.namespace,
                        node.start_point if start_point is None else start_point,
                        node.end_point,
                        )
        self.visit_block(node.children[-1], ctx)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L217
    def visit_expression_statement(self, node: Node, parent: Context):
        check_node(node, "expression_statement")
        left = node.children[0]

        if left.type == "assignment":
            if left.named_child(0).type == "identifier":
                name = node.children[0].children[0].text.decode()
                self.add_symbol("variable", f"{parent.namespace}.{name}")
            return
    #     todo: add more

    def add_symbol(self, kind, id, start=None, end=None):
        self.symbols.append(Symbol(
            kind=kind,
            id=id,
            start=start,
            end=end,
            file_path=self.code_path
        ))



from pathlib import Path

from indexer.scan import ModuleFile, DirPackageScanner
from parser.base import Symbol
from parser.parser import parse_code
import pyarrow.parquet as pq
import pyarrow as pa

import time


def symbol_to_dict(symbol: Symbol):
    start_line, start_col = symbol.start if symbol.start else (None, None)
    end_line, end_col = symbol.end if symbol.end else (None, None)

    return {
        "id": symbol.id,
        "kind": symbol.kind,
        "file_path": symbol.file_path,
        "start_line": start_line,
        "start_col": start_col,
        "end_line": end_line,
        "end_col": end_col
    }




class SymbolIndexer:
    def __init__(self, files: list[ModuleFile], work_path):
        self.work_path = work_path
        self.files = files
        self.symbols: list[Symbol] = []
        self.time_used = 0

    @classmethod
    def from_dir(cls, dir_path):
        scanner = DirPackageScanner(Path(dir_path))
        scanner.scan()
        return cls(scanner.files, dir_path)

    def index(self):
        total = len(self.files)
        cnt = 0
        for file in self.files:
            cnt += 1
            if cnt % 30 == 0:
                print(f"Indexing {cnt}/{total}: {file.name}")
            result = parse_code(file.path, file.name, self.work_path)
            self.symbols.extend(result.symbols)
            self.time_used += result.time_used_ms

        print(f"parsed done, used: {self.time_used}ms")
        print(f"found {len(self.symbols)} symbols")

        start = time.time()
        _symbols = [symbol_to_dict(s) for s in self.symbols]
        table = pa.Table.from_pylist(_symbols,
                                     schema=pa.schema([
                                         pa.field('id', pa.string()),
                                         pa.field('kind', pa.dictionary(pa.int16(), pa.string())),
                                         pa.field('start_line', pa.int32()),
                                         pa.field('start_col', pa.int32()),
                                         pa.field('end_line', pa.int32()),
                                         pa.field('end_col', pa.int32()),
                                        pa.field('file_path', pa.dictionary(pa.int16(), pa.string())),
                                     ]))

        pq.write_table(table, 'symbol_index.parquet',
                       compression='snappy',
                       use_dictionary=True,
                       )


        print(f"writing done, used: {(time.time() - start)*1e3}ms")
        print(f"Indexing done.")



import pyarrow.parquet as pq
import pyarrow as pa

class SymbolSearcher:
    def __init__(self, index_path):
        self.table = pq.read_table(index_path)

    def tokens(self):
        tokens = set()
        ids = self.table.column('id')
        for i in ids:
            parts = i.as_py().split('.')
            tokens.update(parts)
        return tokens

    def search(self, keywords: list[str]):
        indexes = []
        ids = self.table.column('id')
        for i, id_value in enumerate(ids):
            id_str = id_value.as_py()
            if any(keyword in id_str for keyword in keywords):
                indexes.append(i)

        result = []
        for i in indexes:
            id_value = ids[i]
            kind_value = self.table.column('kind')[i]
            start_line = self.table.column('start_line')[i]
            start_col = self.table.column('start_col')[i]
            end_line = self.table.column('end_line')[i]
            end_col = self.table.column('end_col')[i]
            result.append({
                'id': id_value.as_py(),
                'kind': kind_value.as_py(),
                'start_line': start_line.as_py(),
                'start_col': start_col.as_py(),
                'end_line': end_line.as_py(),
                'end_col': end_col.as_py()
            })
        return result






import logging
from typing import List

from pydantic import BaseModel
from tree_sitter import Node

from parser.walker import get_imports
from utils.base import ImportsMap, check_node, Context


class Occurrence(BaseModel):
    id: str
    name: str
    kind: str


class Visitor:
    def __init__(self, node, namespace, imports: ImportsMap):
        self.node = node
        self.namespace = namespace
        self.imports = imports
        self.symbols = {

        }
        self.occurrences = []

    def visit(self, node):
        pass

    def extend_occurrences(self, occurrences: List[Occurrence]):
        self.occurrences.extend(occurrences)


#  _simple_statement: $ => choice(
#       $.future_import_statement,
#       $.import_statement,
#       $.import_from_statement,
#       $.print_statement,
#       $.assert_statement,
#       $.expression_statement,
#       $.return_statement,
#       $.delete_statement,
#       $.raise_statement,
#       $.pass_statement,
#       $.break_statement,
#       $.continue_statement,
#       $.global_statement,
#       $.nonlocal_statement,
#       $.exec_statement,
#       $.type_alias_statement,
#     ),


#  _compound_statement
#       $.if_statement,
#       $.for_statement,
#       $.while_statement,
#       $.try_statement,
#       $.with_statement,
#       $.function_definition,
#       $.class_definition,
#       $.decorated_definition,
#       $.match_statement,

IGNORE_STATEMENTS = [
    'print_statement',
    'assert_statement',
    'return_statement',
    'delete_statement',
    'raise_statement',
    'pass_statement',
    'break_statement',
    'continue_statement',
    'global_statement',
    'nonlocal_statement',
    'exec_statement',
    'type_alias_statement',
]
TODO_STATEMENTS = [
]


# case 'future_import_statement':
#                 names = [i.text.decode() for i in child.named_children]
#                 imports_map.add_import("__future__", names)
#
#             case 'import_statement':
#                 names = [i.text.decode() for i in child.named_children]
#                 for i in names:
#                     imports_map.add_import(i, ["*"])
#
#             case 'import_from_statement':
#                 names = [i.text.decode() for i in child.named_children]
#                 check_node(child.children[0], "from")
#                 check_node(child.children[2], "import")
#                 imports_map.add_import(names[0], names[1:])
def get_future_imports_statement(node):
    check_node(node, "future_import_statement")
    names = [i.text.decode() for i in node.named_children]
    return [
        Occurrence(
            id=f"__future__.{i}",
            name=i,
            kind="ReadAccess",
        )
        for i in names
    ]


def get_imports_statement(node):
    names = [i.text.decode() for i in node.named_children]
    return [
        Occurrence(
            id=f"{i}.__init__",
            name=i,
            kind="ReadAccess",
        )
        for i in names
    ]


def get_import_from_statement(node):
    names = [i.text.decode() for i in node.named_children]
    lib_name = names[0]
    return [
        Occurrence(
            id=f"{lib_name}.{i}",
            name=i,
            kind="ReadAccess",
        )
        for i in names[1:]
    ]


def get_expression_statement(node: Node):
    child = node.children[0]
    match child.type:
        case 'assignment':
            name = child.child(0).text.decode()
            return [
                Occurrence(
                    id=f"{name}",
                    name=name,
                    kind="WriteAccess",
                )
            ]
        case 'augmented_assignment':
            name = child.child(0).text.decode()
            return [
                Occurrence(
                    id=f"{name}",
                    name=name,
                    kind="ReadAccess",
                )
            ]
        case _:
            logging.debug(f"Unsupported expression statement: {child.type}")
            return []


class BlockVisitor(Visitor):

    def visit(self, node):
        for child in node.children:
            if child.type in IGNORE_STATEMENTS:
                continue
            match child.type:
                case 'future_import_statement':
                    self.extend_occurrences(get_future_imports_statement(child))
                case 'import_statement':
                    self.extend_occurrences(get_future_imports_statement(child))
                case 'import_from_statement':
                    self.extend_occurrences(get_future_imports_statement(child))

                case 'print_statement':
                    # todo
                    pass
                case 'assert_statement':
                    # todo
                    pass

                # case 'expression_statement':

                case _:
                    logging.debug(f"Unsupported statement: {child.type}")


class PythonTreeWalker:
    def __init__(self,
                 root_node: Node,
                 content_lines: List[str],
                 module_path: str,
                 ):
        self.root_node = root_node
        self.content_lines = content_lines
        self.module_path = module_path
        self.symbols = []
        self.module_imports = []

    def visit(self):
        assert self.root_node.type == 'module'
        imports = get_imports(self.root_node)
        parent = Context(self.root_node, self.module_path, imports)
        parent.namespace = self.module_path

        self.module_imports = parent.imports.get_libs()
        self.visit_block(self.root_node, parent)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L531

    def visit_decorated_definition(self, node: Node, parent: Context):
        check_node(node, "decorated_definition")

        last_child = node.children[-1]
        match last_child.type:
            case 'function_definition':
                self.visit_function_definition(last_child, parent)
            case 'class_definition':
                self.visit_class_definition(last_child, parent)
            case _:
                raise AssertionError(f"Unsupported decorated definition: {last_child.type}")

    def visit_block(self, node: Node, parent: Context):
        for child in node.children:
            match child.type:
                case 'function_definition':
                    self.visit_function_definition(child, parent)
                case 'class_definition':
                    self.visit_class_definition(child, parent)
                case 'expression_statement':
                    self.visit_expression_statement(child, parent)
                case 'decorated_definition':
                    self.visit_decorated_definition(child, parent)

    # if have decorator, will set start_point
    def visit_function_definition(self, node: Node, parent: Context, start_point=None):

        check_node(node, "function_definition")
        check_node(node.children[0], "def")
        check_node(node.children[1], "identifier")

        func_name = node.children[1].text.decode()
        ctx = Context(node, f"{parent.namespace}.{func_name}", parent.imports)
        self.add_symbol("function",
                        ctx.namespace,
                        node.start_point if start_point is None else start_point,
                        node.end_point,
                        )
        self.visit_block(node, ctx)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L472
    # if have decorator, will set start_point
    def visit_class_definition(self, node: Node, parent: Context, start_point=None):
        check_node(node, "class_definition")
        check_node(node.children[0], "class")
        check_node(node.children[1], "identifier")
        check_node(node.children[-1], "block")

        class_name = node.children[1].text.decode()
        ctx = Context(node, f"{parent.namespace}.{class_name}", parent.imports)
        self.add_symbol("class",
                        ctx.namespace,
                        node.start_point if start_point is None else start_point,
                        node.end_point,
                        )
        self.visit_block(node.children[-1], ctx)

    # https://github.com/tree-sitter/tree-sitter-python/blob/master/grammar.js#L217
    def visit_expression_statement(self, node: Node, parent: Context):
        check_node(node, "expression_statement")
        left = node.children[0]

        if left.type == "assignment":
            check_node(node.children[0].children[0], "identifier")
            name = node.children[0].children[0].text.decode()
            self.add_symbol("const", f"{parent.namespace}.{name}")
            return

    def add_symbol(self, kind, id, start=None, end=None):
        self.symbols.append({
            "kind": kind,
            "id": id,
            "start": start,
            "end": end,
        })



from parser.parser import ParseResult


def print_result(result: ParseResult):
    print(f"Module: {result.module_path}")
    print(f"Path: {result.relative_path}")
    print(f"Time used: {result.time_used_ms}ms")

    print("======== Imports =========")
    for i in result.module_imports:
        print(f"- {i}")

    print("======== Symbols =========")
    for s in result.symbols:
        name = s.id[len(result.module_path) + 1:] if s.id.startswith(result.module_path) else s.id
        print(f"- {s.kind.upper().ljust(10, " ")} :{name} ")


# auto remove all line indent (size=start_col)
def block_lines_between(start_line, start_col, end_line, end_col, lines):
    if start_line == end_line:
        return lines[start_line][start_col:end_col]
    else:
        results = []
        for i in range(start_line, end_line + 1):
            line = lines[i]
            results.append(line)
        return "".join(results)


def print_code_block(result: ParseResult):
    with open(result.relative_path, "r") as f:
        lines = f.readlines()

    for s in result.symbols:
        if s.kind in ["function"]:
            print(f"=========={s.kind}: {s.id} ===============")
            print()
            code = block_lines_between(s.start[0], s.start[1], s.end[0], s.end[1], lines)
            print(code)
            print()



from pydantic import BaseModel


class Symbol(BaseModel):
    id: str
    kind: str
    start: tuple[int, int] | None
    end: tuple[int, int] | None
    file_path: str


class ParseResult(BaseModel):
    module_path: str
    relative_path: str
    time_used_ms: float
    symbols: list[Symbol]
    module_imports: list[str]





class BlockVisitor():
    def __init__(self, node):
        self.node = node

    def visit(self, visitor):
        visitor.visit(self.node)
        for child in self.node.children:
            visitor.visit(child)


import os.path
import time

import tree_sitter_python as tspython
from tree_sitter import Language, Parser, Node

from parser.base import ParseResult
from parser.walker import PythonTreeWalker

PY_LANGUAGE = Language(tspython.language())
parser = Parser(PY_LANGUAGE)


def parse_code(relative_path: str, module_path: str, work_path: str):
    start = time.time()
    with open(os.path.join(work_path, relative_path), "rb") as f:
        content_raw = f.read()
    tree = parser.parse(content_raw)
    content_lines = content_raw.decode().split("\n")

    visitor = PythonTreeWalker(tree.root_node, content_lines, module_path, relative_path)
    visitor.visit()

    return ParseResult(
        module_path=module_path,
        relative_path=relative_path,
        time_used_ms=int((time.time() - start) * 1e6) / 1e3,
        symbols=visitor.symbols,
        module_imports=visitor.module_imports
    )
